/*
 *		l z r w 1 _ c m p r . c
 *
 *    This algorithm appeared in "Practical Evaluation of a Data
 *  Compression Algorithm" in Proceedings Data Compression Conference '91
 *  pp. 362 - 371 by Ross N. Williams of Renaissance Software.
 */

#pragma check_stack(off)
#include <memory.h>

#include "systypes.h"
#include "lzrw1.h"

static char far * far * hash;

void lzrw1_sethash(unsigned long h)
{
	hash = (char far * far *) h;
}
/******************************************************************************/
 
#define TRUE 1
#define PS *p++ != *s++  /* Body of inner unrolled matching loop.         */
#define ITEMMAX 16       /* Maximum number of bytes in an expanded item.  */
// UBYTE *hash[LZRW1_HASHSIZE];

void lzrw1_compress(UBYTE *p_src_first, ULONG src_len, UBYTE *p_dst_first, ULONG *p_dst_len)
{
 UBYTE *p_control;

 /* p_src and p_dst step through the source and destination blocks.           */
 UBYTE *p_src = p_src_first;
 UBYTE *p_dst = p_dst_first;
 
 /* The following variables are never modified and are used in the            */
 /* calculations that determine when the main loop terminates.                */
 UBYTE *p_src_post = p_src_first + src_len;
 UBYTE *p_dst_post = p_dst_first + src_len;
 UBYTE *p_src_max1 = p_src_post - ITEMMAX;
 UBYTE *p_src_max16 = p_src_post - 16*ITEMMAX;

 UWORD control = 0;
 UWORD control_bits = 0;

  *p_dst = FLAG_COMPRESS;
  p_dst += FLAG_BYTES;
  p_control = p_dst;
  p_dst += 2;

  while (TRUE)
  {
    register UBYTE *p;
    register UBYTE *s; 
    UWORD len, index;
    ULONG offset;
    UWORD unroll = 16;

    /* If segment does not compress, do a copy instead */
    if (p_dst > p_dst_post) goto overrun;
    /* If within 16 uncompressed items of the end */
    if (p_src > p_src_max16)
    {
      unroll = 1;
      if (p_src > p_src_max1)
      {
        if (p_src == p_src_post)
	    break;	/* Exit loop once at end of source */
	goto literal;
      }
    }

begin_unrolled_loop:
    /* Compute index from next 3 source characters */
    index = ((40543 * ((((p_src[0]<<4)^p_src[1])<<4)^p_src[2])) >> 4) & 0x0FFF;
    /* Hash the index */                                          
    p = hash[index];
    /* Set new hash table entry to the current pointer */
    hash[index] = s = p_src;
    /* Compute offset between current pointer and hash table entry */
    offset = s - p;

    /* Check that target of hash table pointer is within range and that
       the first three characters of the target match the next three
       characters of the source */
    if (offset > LZRW1_HASHSIZEM1 || p < p_src_first || offset == 0
        || PS || PS || PS)
    {
literal:
      /* If hash table entry is not valid, do a literal copy of the next byte */
      *p_dst++ = *p_src++;
      control >>= 1;	/* Set MSB of control to 0 for a literal */
    }
    else
    {
      /* Compare source and target until a difference is found, up to a
	 maximum of 16 characters (once here, first 3 already matched) */
      PS || PS || PS || PS || PS || PS || PS ||
      PS || PS || PS || PS || PS || PS || s++;
      len = s - p_src - 1;
      /* Code offset and length in 2 bytes */
      *p_dst++ = ((offset & 0x0F00) >> 4) + (len - 1);
      *p_dst++ = offset & 0x00FF;
      p_src += len;
      control = (control>>1) | 0x8000; /* Set MSB of control to 1 for string */
    }
    control_bits++;
    if (--unroll) goto begin_unrolled_loop;

    /* After a full ITEMMAX iterations, set value of control word */
    if (control_bits == ITEMMAX)
    {
      *p_control = control & 0x00FF;
      *(p_control+1) = control >> 8;
      p_control = p_dst;
      p_dst += 2;
      control = control_bits = 0;
    }
  }

  control >>= 16 - control_bits;
  *p_control++ = control & 0x00FF;
  *p_control++ = control >> 8;
  if (p_control == p_dst) p_dst -= 2;
  *p_dst_len = p_dst - p_dst_first;
  return;

overrun:            
  /* If negative compression, do a straight copy */
  *p_dst_first = FLAG_COPY;
  _fmemcpy(p_dst_first+FLAG_BYTES, p_src_first, (unsigned) src_len);
  *p_dst_len = src_len + FLAG_BYTES;
  return;
} 

/******************************************************************************/
void lzrw1_decompress(UBYTE *p_src_first, ULONG src_len, UBYTE *p_dst_first, ULONG *p_dst_len)
{
  UWORD 		controlbits = 0;
  UWORD			control;
  /* Byte pointers p_src and p_dst scan through the input and output blocks.*/
  register UBYTE	*p_src = p_src_first + FLAG_BYTES;
  register UBYTE	*p_dst = p_dst_first;
  UBYTE			*p_src_post = p_src_first + src_len;
   

 /* Check the leading copy flag to see if the compressor chose to use a copy  */
 /* operation instead of a compression operation. If a copy operation was     */
 /* used, then all we need to do is copy the data over, set the output length */
 /* and return.                                                               */
 if (*p_src_first == FLAG_COPY)
 {
    _fmemcpy(p_dst_first, p_src_first + FLAG_BYTES, (unsigned) src_len - FLAG_BYTES);
    *p_dst_len = src_len - FLAG_BYTES;
    return;
 }

 while (p_src != p_src_post)
 {  /* Start of outer loop */
    if (!controlbits)	/* If processed all control bits, get 16 new ones */
    {
      control = *p_src++;
      control |= (*p_src++) << 8;
      controlbits = 16;
    }
    if (control & 1)		/* To copy item, get offset and length */
    {
      UWORD offset, len;
      UBYTE *p;

      offset = (*p_src & 0xF0) << 4;
      len = 1 + (*p_src++ & 0x0F);
      offset += *p_src++ & 0xFF;
      p = p_dst - offset;
      while (len--) *p_dst++ = *p++;
    }
    else
    {
      *p_dst++ = *p_src++;		/* Literal item (one byte) */
    }

    control >>= 1;		/* Shift for next control bit */
    controlbits--;	
 } /* End of outer loop */

 /* Determine the length of the decompressed data before returning. */
 *p_dst_len = p_dst - p_dst_first;
}
