#pragma check_stack(off)
#include <string.h>
#include <stdlib.h>
#include <dos.h>
#ifdef __BORLANDC__
#include <alloc.h>
#include <mem.h>
#else
#include <malloc.h>
#include <memory.h>
#endif

#include "debug.h"
#include "tsr.h"
/*
 * This is an generic module for tsr programming. It provides
 * for routines to go resident and to unload the latest instance.
 * The concept is from "Undocumented Dos" but I rewrote some things
 * to get it as modular as possible.
 *
 * When a tsr loads, it looks for a free entry in the interrupt
 * table. It uses it to store a pointer to a signature structure.
 * This structure is stored in the _psp of our tsr. Upon unloading
 * we scan the interrupt table for a pointer to our signature.
 *
 * Furthermore, I strongly recommend that you use tsr_malloc and
 * tsr_calloc to allocate memory. In this way the modula can keep
 * track off the upper memory address of the data segment. 
 * In "undocumented dos" there was no explanation of an tsr allocting
 * memory; they just assumed it uses static arrays. 
 * I only get this thing working poperly with MS C++ 7.0, Borland
 * messes up; pitty.
 */

signature_t		sigrec = { 8, "", 0, "" };	/* our signature record */


unsigned 		MemTop=0;
unsigned 		Alloc_Size=0;



void *tsr_malloc(unsigned sz)
{
	return tsr_calloc(1,sz);
}

void * tsr_calloc(unsigned items, unsigned item_size)
{
	void 		*ptr;
	unsigned	sz;

	ptr=calloc(items,item_size);
	sz=items*item_size;
	Alloc_Size+=sz;
	MemTop=(((unsigned) ptr)+sz > MemTop)? ((unsigned) ptr)+sz : MemTop;

	return ptr;
}


void tsr_go_resident(char *signame, void *user_data)
{
	uint16 tsr_paras;		// Paragraphs to terminate and leave resident.
	uint16 highest_seg,dummy;

	uint8 far * buf;
	int i;


	if (!signame)
		return ;

	/*
	 * Find ourselves a free interrupt to call our own. Without it,
	 * we can still load, but a future invocation of Phantom with -U
	 * will not be able to unload us.
	 */

	for (i = 0x60; i < 0x67; i++)
	{
		int32 far* p;
		p = (int32 far *) MK_FP(0, ((uint16) i * 4));
		if (*p == 0L)
			break;
	}


	if (i == 0x67)
	{
		tsr_print_string("No user intrs available. TSR is not unloadable..",
			1);
		return;
	}

	/*
	 * Our new found 'user' interrupt will point at the command line area of
	 * our PSP. Complete our signature record, put it into the command line,
	 * then go to sleep.
	 */

	_dos_setvect(i, (INTVECT) (buf = MK_FP(_psp, 0x80)));

	strncpy(sigrec.signature,signame,MAX_SIG);
	memcpy(sigrec.user_data,user_data,MAX_USR);
	sigrec.psp = _psp;
	*((signature_t far *) buf) = sigrec;



	_asm mov highest_seg, ds;

	tsr_paras= highest_seg + 1 + ( (MemTop+16) >>4) - _psp;
	dprintf(D_SYSTEM, "Alloc_Size:%u, Memtop:%u, tsr_paras*16:%u\n", Alloc_Size,MemTop,(tsr_paras<<4));


	_dos_setblock(tsr_paras, _psp,&dummy);
	_dos_keep(0, tsr_paras);

	// we're gone....
}

void tsr_unload_latest(char *name, unload_callbackfunc unload_func)
{
	int 				i;
	INTVECT 			p_vect;
	signature_t far *	sig_ptr;
	uint16 				psp;
	static uint16 		save_ss, save_sp;

	/*
	 * Note that we step backwards to allow unloading of Multiple copies
	 * in reverse order to loading, so that the Int 2Fh chain remains
	 * intact.
	 */
	for (i = 0x66; i >= 0x60; i--)
	{
		int32 far* p;
		p = (int32 far *) MK_FP(0, ((uint16) i * 4));
		sig_ptr = (signature_t far *) *p;
		if (_fstrncmp(sig_ptr->signature, name,MAX_SIG) == 0)
			break;
	}

	if (i == 0x5f)
	{
		tsr_print_string(name,0);
		tsr_print_string(" was not loaded.",1);
		return;
	}

	if (unload_func && !unload_func(sig_ptr))
	{
		tsr_print_string("TSR is not unloaded.",1);
		return;
	}

	p_vect = _dos_getvect(i);
	psp = ((signature_t far *) p_vect)->psp;

	/*
	 * Use the recommended switch PSP and Int 4Ch method of
	 * unloading the TSR (see TSRs chapter of Undocumented DOS).
	 */

	_asm {
		// Save some registers
		push	es;
		push	ds;
		push	si;
		push	di;
		push	bp;
		// Set resident program's parent PSP to us.
		mov		es, psp;
		mov		bx, 0x16;
		mov		ax, _psp;
		mov		es:[di], ax;
		mov		di, 0x0a;
		// Set resident program PSP return address to exit_ret;
		mov		ax, offset exit_ret;
		stosw;
		mov		ax, cs;
		stosw;
		mov		bx, es;
		// Set current PSP to resident program
		mov		ah, 0x50;
		int		21h;
		// Save SS:SP
		mov		ax, seg save_ss;
		mov		ds, ax;
		mov		save_ss, ss;
		mov		save_sp, sp;
		// and terminate
		mov		ax, 0x4c00;
		int		21h;
	}
exit_ret:
	_asm {
		// We should arrive back here - restore SS:SP
		mov		ax, seg save_ss;
		mov		ds, ax;
		mov		ss, save_ss;
		mov		sp, save_sp;
		// restore the registers
		pop		bp;
		pop		di;
		pop		si;
		pop		ds;
		pop		es;
		// Set current PSP back to us.
		mov		bx, _psp;
		mov		ah, 0x50;
		int		21h;
	}

	_dos_setvect(i, NULL);
	tsr_print_string("TSR unloaded succesfully.",1);
}

static void print_char(char c)
{
	_asm {
		mov ah, 0eh;
		mov al, byte ptr c;
		int 10h;	
	}
}

void tsr_print_string(char far* str, int add_newline)
{
	while (*str) print_char(*str++);
	if (add_newline)
	{
		print_char('\n');
		print_char('\r');
	}
}
